# Created as a part of Master's Thesis "Using machine learning methods to save energy in a smart home"
# Faculty of Information Technology, Brno University of Technology, 2024
# Author: Adam Grünwald
#
# This module is used for controlling the boiler with use of the predictions in combination with the smart heating algotithm.


from pathlib import Path
from matplotlib.font_manager import json_load
import pytz
from datetime import datetime, timedelta

import os.path
import logging
import time
import os
import json
from dateutil import parser



from smartboiler.event_checker import EventChecker
from smartboiler.data_handler import DataHandler
from smartboiler.fotovoltaics import Fotovoltaics
from smartboiler.forecast import Forecast
from smartboiler.boiler import Boiler


class Controller:
    """Main class which makes decisions about about heating"""

    def __init__(
        self,
        dataHandler: DataHandler,
        boiler: Boiler,
        forecast: Forecast,
        eventChecker: EventChecker,
        load_model=False,
        learning=True,
    ):
        """Inits class of Controller. Loads settings from a settings file

        Args:
            settings_file (str, optional): [name of json file with settings]. Defaults to 'settings.json'.
        """

        self.tmp_min = 5
        self.learning = learning

        self.start_date = datetime.now()

        self.dataHandler = dataHandler
        self.boiler = boiler
        self.forecast = forecast
        self.eventChecker = eventChecker

        if load_model:
            print("loading model")
            self.forecast.build_model()
            self.forecast.load_model()
        else:
            print("training model")
            self.forecast.build_model()
            forecast.train_model()

        self.last_model_training = datetime.now()
        self.last_legionella_heating = datetime.now()

    def _last_entry(self):
        print("getting last entry")
        last_entry = self.dataHandler.get_actual_boiler_stats()
        print("last entry: {}".format(last_entry))
        return last_entry

    def _check_data(self):
        """Retrain model if the last data update is older than 7 days."""
        if self.last_model_training - datetime.now() > timedelta(days=7):
            print("actualizing data")

            self.forecast.train_model()
            self.last_model_training = datetime.now()

    def _learning(self):
        """After one week of only measuring the data starts heating based on historical data.

        Returns:
            [boolean]: [True if in learing]
        """
        if not self.learning:
            return False
        return (datetime.now() - self.start_date) < timedelta(days=28)

    def actualize_forecast(self):
        self.actual_forecast = self.forecast.get_forecast_next_steps()

    def control(self):
        """Method which decides about turning on or off the heating of a boiler."""

        time_now = datetime.now().astimezone(pytz.timezone("Europe/Prague"))
        print(f"actual time: {time_now}, controling boiler")
        last_entry = self._last_entry()

        # checks whether the water in boiler should be even ready
        if self.eventChecker.check_off_event():
            print("turning off boiler, event in calendar")
            self.boiler.turn_off()
            return
        # check scheduled heating target event

        tmp_measured = last_entry["boiler_case_tmp"]
        is_on = last_entry["is_boiler_on"]

        # actual tmp of water in boiler
        tmp_act = self.boiler.real_tmp(tmp_measured)
        print(f"actual tmp: {tmp_act}, measured: {tmp_measured}")

        if is_on is None:
            print("boiler state is unknown")
            is_on = False

        # protection from freezing
        if tmp_act < 5:
            self.boiler.turn_on()

        if self._learning():
            if tmp_act > 60:
                if is_on:
                    self.boiler.turn_off()
            else:
                if tmp_act < 57:
                    if not is_on:
                        self.boiler.turn_on()
            return
        else:
            self._check_data()

        # if the boiler is needed to heat up before the next predicted consumption
        is_needed_to_heat, minutes_needed_to_heat = self.boiler.is_needed_to_heat(
            tmp_act, self.actual_forecast
        )
        if is_needed_to_heat:
            print(
                f"need to heat for {minutes_needed_to_heat} minutes, turning on and sleeping"
            )
            self.boiler.turn_on()
            time.sleep(minutes_needed_to_heat * 60)
        else:
            print("no need to heat, turning off")
            self.boiler.turn_off()

        if self.last_legionella_heating - datetime.now() > timedelta(days=21):
            self.boiler.turn_on()
            if tmp_act >= (65):
                time.sleep(1200)
                self.last_legionella_heating = datetime.now()
                self.boiler.turn_off()
                return


if __name__ == "__main__":
    logging.info("Starting SmartBoiler Controller")

    OPTIONS_PATH = os.getenv("OPTIONS_PATH", default="/app/options.json")
    options_json = Path(OPTIONS_PATH)

    # Read options info
    if options_json.exists():
        with options_json.open("r") as data:
            options = json.load(data)

    DATA_PATH = os.getenv("DATA_PATH", default="/app/data/")

    data_path = Path(DATA_PATH)


    shelly_ip = options["shelly_ip"]
    boiler_switch_id = options["boiler_switch_id"]
    
    start_of_data_measurement = parser.parse(options["start_of_data_measurement"])

    home_longitude = options["home_longitude"]
    home_latitude = options["home_latitude"]
    device_tracker_entity_id = options["device_tracker_entity_id"]
    device_tracker_entity_id_2 = options["device_tracker_entity_id_2"]

    influxdb_host = options["influxdb_host"]
    influxdb_port: 8086
    influxdb_user = options["influxdb_user"]
    influxdb_pass = options["influxdb_pass"]
    influxdb_name = options["influxdb_name"]

    boiler_case_tmp_entity_id = options["boiler_case_tmp_entity_id"]
    boiler_water_flow_entity_id = options["boiler_water_flow_entity_id"]
    boiler_water_temp_entity_id = options["boiler_water_temp_entity_id"]
    boiler_water_temp_entity_id_2 = options["boiler_water_temp_entity_id_2"]
    boiler_volume = options["boiler_volume"]
    boiler_set_tmp = options["boiler_set_tmp"]
    boiler_min_operation_tmp = options["boiler_min_operation_tmp"]
    average_boiler_surroundings_temp = options["average_boiler_surroundings_temp"]
    boiler_case_max_tmp = options["boiler_case_max_tmp"]
    boiler_watt_power = options["boiler_watt_power"]

    logging_level = options["logging_level"]

    load_model = options["load_model"]
    model_type = options["model_type"]
    learning = options["learning"]

    hdo = options["hdo"]
    has_fotovoltaics = options["has_fotovoltaics"]
    fve_solax_sn = options["fve_solax_sn"]
    fve_solax_token = options["fve_solax_token"]

    # chosing the model based on size of household
    if model_type == "smaller_household":
        model_path = "/app/model_form.weights.h5"
        scaler_path = "/app/scaler_form.pkl"
    else:
        model_path = "/app/model_zuka.weights.h5"
        scaler_path = "/app/scaler_zuka.pkl"
        
    model_path = Path(model_path)
    scaler_path = Path(scaler_path)

    dataHandler = DataHandler(
        influx_id=influxdb_host,
        db_name=influxdb_name,
        db_username=influxdb_user,
        db_password=influxdb_pass,
        relay_entity_id=boiler_switch_id,
        tmp_boiler_case_entity_id=boiler_case_tmp_entity_id,
        tmp_output_water_entity_id=boiler_water_temp_entity_id,
        tmp_output_water_entity_id_2=boiler_water_temp_entity_id_2,
        device_tracker_entity_id=device_tracker_entity_id,
        device_tracker_entity_id_2=device_tracker_entity_id_2,
        home_longitude=home_longitude,
        home_latitude=home_latitude,
        start_of_data=start_of_data_measurement,
    )

    print("inicializing fotovoltaics from controller __main__")
    if has_fotovoltaics:
        fotovoltaics = Fotovoltaics(
            token=fve_solax_token,
            sn=fve_solax_sn,
        )
    else:
        fotovoltaics = None
    eventChecker = EventChecker()
    print("inicializing boiler from controller __main__")
    boiler = Boiler(
        shelly_ip,
        dataHandler=dataHandler,
        eventChecker=eventChecker,
        fotovoltaics=fotovoltaics,
        capacity=boiler_volume,
        wattage=boiler_watt_power,
        set_tmp=boiler_set_tmp,
        min_tmp=boiler_min_operation_tmp,
        average_boiler_surroundings_temp=average_boiler_surroundings_temp,
        boiler_case_max_tmp=boiler_case_max_tmp,
        hdo=hdo,
    )

    predicted_columns = ["longtime_mean"]
    print("inicializing forecast from controller __main__   ")
    forecast = Forecast(
        dataHandler=dataHandler,
        start_of_data=start_of_data_measurement,
        model_path=model_path,
        scaler_path=scaler_path,
        predicted_columns=predicted_columns,
    )
    print("inicializing controller from controller __main__")
    controller = Controller(
        dataHandler=dataHandler,
        boiler=boiler,
        forecast=forecast,
        eventChecker=eventChecker,
        load_model=load_model,
        learning=learning,
    )

    while 1:
        try:
            controller.actualize_forecast()
            for i in range(0, 15):
                controller.control()
                time.sleep(60)
        except Exception as e:
            print(f"Error: {e}")
            time.sleep(60)



# Created as a part of Master's Thesis "Using machine learning methods to save energy in a smart home"
# Faculty of Information Technology, Brno University of Technology, 2024
# Author: Adam Grünwald
#
# This module is used for communicating with Google Calendar API and searching for special events.


from __future__ import print_function
from datetime import datetime, timezone, timedelta
import pickle
import os.path
import re

from google.auth.transport.requests import Request
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import build
from google.oauth2.credentials import Credentials

from smartboiler.time_handler import TimeHandler


class EventChecker:
    """Class for communicating with Google Calendar API and searching for special events."""

    def __init__(self):
        """Initialize the class with the scopes for Google Calendar API and TimeHandler class."""
        self.SCOPES = ["https://www.googleapis.com/auth/calendar.readonly"]
        self.TimeHandler = TimeHandler()

    def load_events(self) -> list:
        """Loads events from Google Calendar API using credentials of Google Calendar.

        Code for communicating with Google Calendar API is inspired by a manual on https://developers.google.com/calendar/
        quickstart/python

        Returns:
            [list]: [list of events]
        """
        try:

            creds = None
            token_file = "/app/token.json"
            creds_file = "/app/credentials.json"

            if os.path.exists(token_file):
                creds = Credentials.from_authorized_user_file(token_file, self.SCOPES)
            # If there are no (valid) credentials available, let the user log in.
            if not creds or not creds.valid:
                if creds and creds.expired and creds.refresh_token:
                    creds.refresh(Request())
                else:
                    flow = InstalledAppFlow.from_client_secrets_file(
                        creds_file, self.SCOPES
                    )
                    creds = flow.run_local_server(port=0)
                # Save the credentials for the next run
                with open(token_file, "w") as token:
                    token.write(creds.to_json())

            service = build("calendar", "v3", credentials=creds)

            now = datetime.now().isoformat() + "Z"

            events_result = (
                service.events()
                .list(
                    calendarId="primary",
                    timeMin=now,
                    maxResults=1,
                    singleEvents=True,
                    orderBy="startTime",
                )
                .execute()
            )
            events = events_result.get("items", [])
            return events
        except Exception as e:
            print("Error while loading events")
            print(e)
            return None

    def next_calendar_heat_up_event(self) -> dict:
        """Search next event in a calendar which contains specific words describing the process of heating up.

        Args:

        Returns:
            [dict]: [dictionary describing next heating up event]
        """
        events = self.load_events()
        return_dict = {"minutes_to_event": None, "degree_target": None}

        if events:
            for e in events:
                if re.match("^.*boiler heat up at (\d+) degrees$", e["summary"]):
                    degree_target = int(
                        re.split("^.*boiler heat up at (\d+) degrees$", e["summary"])[1]
                    )
                    start = self.TimeHandler.date_to_datetime(
                        e["start"].get("dateTime", e["start"].get("date"))
                    )
                    end = self.TimeHandler.date_to_datetime(
                        e["end"].get("dateTime", e["end"].get("date"))
                    )
                    time_to_event = (
                        start - (datetime.now() + timedelta(hours=1))
                    ) / timedelta(hours=1)
                    time_to_end_event = (
                        end - (datetime.now() + timedelta(hours=1))
                    ) / timedelta(hours=1)

                    if time_to_event > 0:

                        return_dict["minutes_to_event"] = time_to_event * 60
                        return_dict["degree_target"] = degree_target

                        return return_dict
                    # case in event
                    elif time_to_end_event > 0:
                        return_dict["minutes_to_event"] = 0
                        return_dict["degree_target"] = degree_target

                        return return_dict

        return return_dict

    def check_off_event(self) -> bool:
        """Search for an event for turning off the boiler.

        Returns:
            [bool]: [if the turn offevent is happening now or not]
        """
        events = self.load_events()
        if events:
            for e in events:

                if "#off" in e["summary"]:
                    start = self.TimeHandler.date_to_datetime(
                        e["start"].get("dateTime", e["start"].get("date"))
                    )
                    end = self.TimeHandler.date_to_datetime(
                        e["end"].get("dateTime", e["end"].get("date"))
                    )
                    return self.TimeHandler.is_date_between(start, end)
        return False


if __name__ == "__main__":
    e = EventChecker()
    print(e.next_calendar_heat_up_event())



"""A setuptools based setup module.

See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / "README.md").read_text(encoding="utf-8")

# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
    name="smartboiler",  # Required
    version="0.1.11",  # Required
    description="Smart boiling of household",  # Optional
    long_description=long_description,  # Optional
    long_description_content_type="text/markdown",  # Optional (see note above)
    url="https://github.com/grinwi/smartboiler",  # Optional
    author="Adam GRUNWALD",  # Optional
    author_email="grunwald.adam24@gmail.com",  # Optional
    classifiers=[  # Optional
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent",
    ],
    keywords="energy, management, optimization, hass",  # Optional
    package_dir={"": "src"},  # Optional
    packages=find_packages(where="src"),  # Required
    python_requires=">=3.9, <=3.10",
    install_requires=[
        "wheel",
        "numpy>=1.22.2",
        "h5py==3.10.0",
        "scipy<1.9.0",
        "pandas>=1.4.1",
        "pvlib>=0.10.1",
        "protobuf>=3.0.0",
        "pytz>=2021.1",
        "requests>=2.25.1",
        "beautifulsoup4>=4.9.3",
        "pulp>=2.4",
        "tensorflow",
        "keras",
        "Mako==1.3.2",
        "pyyaml>=5.4.1",
        "tables==3.7.0",
        "skforecast==0.10.1",
        "influxdb==5.3.1",
        "matplotlib==3.5.1",
        "flask>=2.0.2",
    ],
    # Optional
    entry_points={  # Optional
        "console_scripts": [
            "smartboiler=smartboiler.command_line:main",
        ],
    },
    package_data={"smartboiler": ["templates/index.html", "static/style.css"]},
)
